"""Monte Carlo method for ase."""
from typing import Dict, Union, Iterator
import sys
import datetime
import time
import logging
import random
import math
from collections import Counter
from ase import Atoms
from ase.units import kB
from clease.version import __version__
from clease.datastructures import SystemChanges, MCStep
from .mc_evaluator import CEMCEvaluator, MCEvaluator
from .base import BaseMC
from .averager import Averager
from .bias_potential import BiasPotential
from .observers import MCObserver
from .trial_move_generator import TrialMoveGenerator, RandomSwap

logger = logging.getLogger(__name__)


# pylint: disable=too-many-instance-attributes
class Montecarlo(BaseMC):
    """Class for running Monte Carlo at a fixed composition (canonical).
    For more information, also see the documentation of the parent class
    :class:`~clease.montecarlo.base.BaseMC`.

    Args:
        system (Union[ase.Atoms, MCEvaluator]): Either an ASE Atoms object
            with an attached calculator, or a pre-initialized
            :class:`~clease.montecarlo.mc_evaluator.MCEvaluator`
            object.
        temp (float): Temperature of Monte Carlo simulation in Kelvin
        generator (TrialMoveGenerator, optional): A
            :class:`~clease.montecarlo.trial_move_generator.TrialMoveGenerator`
            object that produces trial moves. Defaults to None.
    """

    NAME = "MonteCarlo"

    def __init__(
        self,
        system: Union[Atoms, MCEvaluator],
        temp: float,
        generator: TrialMoveGenerator = None,
    ):
        # We cannot cause an energy calculation trigger in init,
        # so we defer these quantities until needed.
        self.current_energy = None
        self.new_energy = None
        self.mean_energy = None
        self.energy_squared = None

        super().__init__(system, temp)

        if generator is None:
            self.generator = _make_default_swap_generator(self.evaluator)
        else:
            self.generator = generator

        # List of observers that will be called every n-th step
        # similar to the ones used in the optimization routines
        self.observers = []
        self.bias_potentials = []

        self.current_step = 0
        self.num_accepted = 0
        self.status_every_sec = 30

        self.trial_move = []

        self.quit = False

    def _on_temp_change(self) -> None:
        """Reset the energy averagers after a change in temperature"""
        self.reset_averagers()

    def update_current_energy(self) -> None:
        self.current_energy = self.evaluator.get_energy()
        logger.debug("Updating current energy to %s", self.current_energy)
        self.evaluator.reset()

    def _initialize_energies(self) -> None:
        """Initialize the energy quantities, causes a trigger of an energy evaluation."""

        self.update_current_energy()
        self.new_energy = self.current_energy

        # Create new averager objects, only the first time
        # (since they are initialized as None in __init__)
        if self.mean_energy is None:
            self.mean_energy = Averager(ref_value=self.current_energy)
        if self.energy_squared is None:
            self.energy_squared = Averager(ref_value=self.current_energy)

    def reset(self) -> None:
        """Reset all member variables to their original values."""
        logger.debug("Resetting.")
        for _, obs in self.observers:
            obs.reset()

        self.evaluator.reset()
        self._reset_internal_counters()
        self.reset_averagers()

    def reset_averagers(self) -> None:
        """Reset the energy averagers."""
        # Averagers are initialized to None in the constructor.
        for averager in (self.mean_energy, self.energy_squared):
            if averager is not None:
                averager.clear()

    def _reset_internal_counters(self) -> None:
        """Reset the step counters which are used internally"""
        self.current_step = 0
        self.num_accepted = 0

    def add_bias(self, potential: BiasPotential):
        """Add a new bias potential.

        Parameters:

        potential:
            Potential to be added
        """
        if not isinstance(potential, BiasPotential):
            raise TypeError("potential has to be of type BiasPotential")
        self.bias_potentials.append(potential)

    def attach(self, obs: MCObserver, interval: int = 1):
        """Attach observers to be called on a given MC step interval.

        Parameters:

        obs: MCObserver
            Observer to be added

        interval: int
            How often the observer should be called
        """
        if not obs.interval_ok(interval):
            name = type(obs).__name__
            raise ValueError(f"Invalid interval for {name}. Check docstring of the observer.")

        self.observers.append((interval, obs))

    def initialize_run(self):
        """Prepare MC object for a new run."""
        logger.debug("Initializing run")
        self.generator.initialize(self.atoms)

        # Ensure the evaluator is properly synchronized.
        self.evaluator.synchronize()

        # Initialize/update relevant energy quantities
        self._initialize_energies()

        # Reset the internal step counters
        self._reset_internal_counters()

    def run(self, steps: int = 100) -> None:
        """Run Monte Carlo simulation.

        Parameters:

        steps: int
            Number of steps in the MC simulation
        """

        # Construct the iterator, make the preparations for starting the run
        mc_iter = self.irun(steps)

        start = time.perf_counter()
        prev = self.current_step
        info_enabled = logger.isEnabledFor(logging.INFO)  # Do we emit INFO logs?
        for _ in mc_iter:
            # We only want to do this calculation if logging is enabled for INFO.
            if info_enabled and time.perf_counter() - start > self.status_every_sec:
                ms_per_step = 1000.0 * self.status_every_sec / (self.current_step - prev)
                accept_rate = self.num_accepted / self.current_step
                logger.info(
                    "%d of %d steps. %.2f ms per step. Acceptance rate: %.2f",
                    self.current_step,
                    steps,
                    ms_per_step,
                    accept_rate,
                )
                prev = self.current_step
                start = time.perf_counter()

    def irun(self, steps: int) -> Iterator[MCStep]:
        """Run Monte Carlo simulation as an iterator.
        Can be used to inspect the MC after each step, for example,
        to print the energy every 5 steps, one could do:

        >>> mc = Montecarlo(...)  # doctest: +SKIP
        >>> for mc_step in mc.irun(500):  # doctest: +SKIP
        ...     if mc_step.step % 5 == 0:
        ...         print(f"Current energy: {mc_step.energy:.2f} eV")

        The iterator yields individual instances of :class:`~clease.datastructures.mc_step.MCStep`
        for each step which is taken.

        Parameters:

        steps: int
            Number of steps in the MC simulation
        """
        logger.info("Starting MC run with %d steps.", steps)

        # We first ensure the MC object is initialized, and then we
        # construct the iterator.
        # This ensures everything is prepared before the first step
        # is consumed from the iterator.
        self.initialize_run()

        # Now we create the iterator
        return self._irun(steps)

    def _irun(self, steps: int) -> Iterator[MCStep]:
        """Create the MC iterator"""
        # Offset range by 1, so that we start with current step = 1
        for _ in range(steps):
            step = self._mc_step()

            en = step.energy
            self.mean_energy += en
            # E * E is slightly faster than E ** 2
            self.energy_squared += en * en

            if self.quit:
                # Breaking out will cause the iterator to end.
                logger.warning("Quit signal detected. Breaking the MC loop.")
                break
            # Yield, allow inspection before continuing.
            yield step
        else:
            # Loop was not broken, we reached the max number of steps
            logger.info("Reached maximum number of steps (%d mc steps)", steps)
        logger.debug("Reached end of MC iterator.")

    @property
    def meta_info(self):
        """Return dict with meta info."""
        ts = time.time()
        st = datetime.datetime.fromtimestamp(ts).strftime("%Y-%m-%d %H:%M:%S")
        clease_version = str(__version__)
        v_info = sys.version_info
        meta_info = {
            "timestamp": st,
            "python_version": f"{v_info.major}.{v_info.minor}.{v_info.micro}",
            "clease_version": clease_version,
        }
        return meta_info

    def get_thermodynamic_quantities(self):
        """Compute thermodynamic quantities."""
        quantities = {}
        mean_energy = self.mean_energy.mean
        quantities["energy"] = mean_energy
        mean_sq = self.energy_squared.mean
        quantities["heat_capacity"] = (mean_sq - mean_energy**2) / (kB * self.temperature**2)
        quantities["energy_var"] = mean_sq - mean_energy**2
        quantities["temperature"] = self.temperature
        at_count = self.count_atoms()
        for key, value in at_count.items():
            name = f"{key}_conc"
            conc = float(value) / len(self.atoms)
            quantities[name] = conc

        # Add some more info that can be useful
        quantities.update(self.meta_info)

        # Add information from observers
        for obs in self.observers:
            quantities.update(obs[1].get_averages())
        return quantities

    def _calculate_step(self, system_changes: SystemChanges):
        """Calculate energies given a step, and decide if we accept the step.

        Returns boolean if system changes are accepted.

        Parameters:

        system_changes: list
            List with system changes
        """
        logger.debug("Applying changes to the system.")
        self.evaluator.apply_system_changes(system_changes)
        # Evaluate the energy quantity after applying the changes to the system.
        self.new_energy = self.evaluator.get_energy(applied_changes=system_changes)

        # NOTE: As this is called after calculate, the changes has
        # already been introduced to the system
        for bias in self.bias_potentials:
            self.new_energy += bias(system_changes)
        logger.debug(
            "Current energy: %.3f eV, new energy: %.3f eV",
            self.current_energy,
            self.new_energy,
        )
        accept = self._do_accept(self.current_energy, self.new_energy)
        logger.debug("Change accepted? %s", accept)

        if accept:
            # Changes accepted, finalize evaluator.
            self.evaluator.keep_system_changes(system_changes)
        else:
            # Undo changes
            self.evaluator.undo_system_changes(system_changes)

        return accept

    def _do_accept(self, current_energy: float, new_energy: float) -> bool:
        """Decide if we accept a state, based on the energies.

        Return a bool on whether the move was accepted.

        Parameters:

        :param current_energy: Energy of the current configuration
        :param new_energy: Energy of the new configuration.
        """
        # Standard Metropolis acceptance criteria
        if new_energy < current_energy:
            return True
        kT = self.temperature * kB
        energy_diff = new_energy - current_energy
        probability = math.exp(-energy_diff / kT)
        logger.debug(
            "Energy difference: %.3e. Calculated probability: %.3f",
            energy_diff,
            probability,
        )
        return random.random() <= probability

    def _move_accepted(self, system_changes: SystemChanges) -> None:
        logger.debug("Move accepted, updating things")
        self.num_accepted += 1
        self.generator.on_move_accepted(system_changes)
        self.current_energy = self.new_energy

    def _move_rejected(self, system_changes: SystemChanges) -> None:
        logger.debug("Move rejected, undoing system changes: %s", system_changes)
        self.generator.on_move_rejected(system_changes)

    def count_atoms(self) -> Dict[str, int]:
        """Count the number of each element."""
        return dict(Counter(self.atoms.symbols))

    def _mc_step(self) -> MCStep:
        """Make one Monte Carlo step by swithing two atoms."""
        self.current_step += 1
        system_changes = self.generator.get_trial_move()
        self.trial_move = system_changes

        # Calculate step, and whether we accept it
        move_accepted = self._calculate_step(system_changes)

        updater = self._move_accepted if move_accepted else self._move_rejected
        updater(system_changes)

        step = MCStep(self.current_step, self.current_energy, move_accepted, system_changes)
        # Execute all observers
        self.execute_observers(step)

        return step

    def execute_observers(self, last_step: MCStep):
        for interval, obs in self.observers:
            if self.current_step % interval == 0:
                logger.debug(
                    "Executing observer %s at step %d with interval %d.",
                    obs.name,
                    self.current_step,
                    interval,
                )
                obs.observe_step(last_step)


def _make_default_swap_generator(evaluator: MCEvaluator) -> RandomSwap:
    """Construct the default RandomSwap trial move generator.
    If the evaluator object is not a Cluster Expansion evaluator,
    then any swaps are allowed.

    If the evaluator is a Cluster Expansion MC evaluator object,
    then only non-background sites are considered for random swaps.
    This does *not* constrain swaps between sublattices.
    """
    atoms = evaluator.atoms
    if not isinstance(evaluator, CEMCEvaluator):
        # Running a non-CE Evaluator
        return RandomSwap(atoms)
    # We're a CE evaluator, constrain the background sites
    non_bkg_indices = evaluator.settings.non_background_indices
    return RandomSwap(atoms, indices=non_bkg_indices)



from typing import Sequence, Set, Dict, List, Iterator, Tuple
import logging
from itertools import product
import functools
from copy import deepcopy
import numpy as np
import ase
from ase.geometry import wrap_positions

from clease import tools
from clease.datastructures import FourVector, Figure, TransMatrix

from .cluster_fingerprint import ClusterFingerprint
from .cluster import Cluster
from .cluster_list import ClusterList
from .cluster_generator import ClusterGenerator
from .utils import name_clusters, size

__all__ = ("ClusterManager",)

logger = logging.getLogger(__name__)


class ClusterManager:
    """
    Manager for construction of all cluster.

    Parameters:

    prim_cell: ase.Atoms
        Primitive cell
    """

    def __init__(self, prim_cell: ase.Atoms, background_syms: Set[str] = None):
        self._background_syms = background_syms or set()

        primitive_filtered = self._filter_background(prim_cell)

        self.generator = ClusterGenerator(primitive_filtered)
        self.clusters = ClusterList()
        self._cache = _CacheChecker()
        # Flag for determining if we can do the "trivial" path in trans_matrix
        # Should only be disabled for testing purposes!
        # Generally, should be set to True".
        self._allow_trivial_path = True

    @property
    def prim(self) -> ase.Atoms:
        """The primitive cell.
        Note, that background atoms have been removed from this atoms object."""
        return self.generator.prim

    @property
    def background_syms(self) -> Set[str]:
        """The symbols which are considered background."""
        return self._background_syms

    def _filter_background(self, atoms: ase.Atoms) -> ase.Atoms:
        """Filter the background atoms from an ASE Atoms object.

        Returns a copy of the original atoms object.
        """
        atoms = atoms.copy()

        # Find the indices we need to delete
        # If no background atoms are present, this will do nothing.
        delete = [atom.index for atom in atoms if self.is_background_atom(atom)]
        delete.sort(reverse=True)
        for i in delete:
            del atoms[i]
        return atoms

    def is_background_atom(self, atom: ase.Atom) -> bool:
        """Check whether an atom is a background atom."""
        return atom.symbol in self.background_syms

    def __eq__(self, other):
        return self.clusters == other.clusters and self.generator == other.generator

    def _prepare_new_build(self, max_cluster_dia):
        """Prepare for a new call to ``build``"""
        # Update the cache
        self._cache.set_cache(max_cluster_dia)
        # Clear any old clusters
        self.clusters.clear()

    def requires_build(self, max_cluster_dia: Sequence[float]) -> bool:
        """Check if the cluster manager requires a new build
        for a given set of cluster diameters"""
        return self._cache.requires_build(max_cluster_dia)

    def build(self, max_cluster_dia: Sequence[float]) -> None:
        """
        Construct all clusters.

        Parameters:

        max_cluster_dia: sequence of floats
            Maximum distance between two atoms in a cluster,
            for each cluster body. The diameters start at 2-body clusters
        """
        # Check if we already built the clusters with these settings
        if not self.requires_build(max_cluster_dia):
            return
        # We got a new set of settings, prepare to construct new clusters
        self._prepare_new_build(max_cluster_dia)

        # Number of lattices from 0 to N, where N
        # is the number of atoms in the primitive,
        # possibly without the background atoms.
        num_lattices = range(len(self.prim))
        all_fps: List[ClusterFingerprint] = []
        all_figures: List[List[Figure]] = []
        lattices = []

        for ref_lattice, (indx, diameter) in product(num_lattices, enumerate(max_cluster_dia)):
            cluster_size = indx + 2  # Size of cluster, start with 2-body at index 0
            figures, fps = self.generator.generate(cluster_size, diameter, ref_lattice)

            all_fps += fps
            all_figures += figures
            lattices += [ref_lattice] * len(figures)

        names = self._get_names(all_fps)
        # Transfer to the cluster list
        for figures, fp, name, ref_lattice in zip(all_figures, all_fps, names, lattices):
            # All figures are of the same size
            cluster_size = figures[0].size
            # Calculate the diameter from the first Figure, since they are all
            # geometrically equivalent, and thus have the same diameter.
            diameter = figures[0].get_diameter(self.prim)
            eq_sites = self.generator.equivalent_sites(figures[0])

            cluster = Cluster(
                name=name,
                size=cluster_size,
                diameter=diameter,
                fingerprint=fp,
                figures=figures,
                equiv_sites=eq_sites,
                group=ref_lattice,
            )
            self.clusters.append(cluster)

        # Add singlets
        for i in range(len(self.prim)):
            self.clusters.append(
                Cluster(
                    name="c1",
                    size=1,
                    diameter=0.0,
                    fingerprint=ClusterFingerprint([1.0]),
                    figures=[Figure([FourVector(0, 0, 0, i)])],
                    equiv_sites=[],
                    group=i,
                )
            )
        # Add empty
        self.clusters.append(
            Cluster(
                name="c0",
                size=0,
                diameter=0.0,
                fingerprint=ClusterFingerprint([0.0]),
                figures=[],
                equiv_sites=[],
                group=0,
            )
        )
        # Put the clusters in order of size. Has no practical effect,
        # but it looks nicer upon inspection.
        self.clusters.sort()

    @staticmethod
    def _get_names(all_fps: Sequence[ClusterFingerprint]):
        """
        Give a consistent name to all clusters

        Parameter:

        all_fps: list of ClusterFingerPrint
            A list with all the cluster fingerprints
        """
        # The following one-liner will do the job when we
        # decide to update cluster names from the legacy ones
        # where names where per by size
        # names = name_clusters(all_fps)

        sizes = sorted(set(size(fp) for fp in all_fps))
        names = [None for _ in all_fps]
        for s in sizes:
            fps = []
            indices = []
            for i, fp in enumerate(all_fps):
                if size(fp) == s:
                    indices.append(i)
                    fps.append(fp)
            names_per_size = name_clusters(fps)
            for i, n in zip(indices, names_per_size):
                names[i] = n
        return names

    def cluster_list_for_template(self, template: ase.Atoms) -> ClusterList:
        """
        Specialise the cluster information to a template. Returns a new
        ClusterList instance.

        Parameter:

        template: ase.Atoms
            Atoms object representing the simulation cell
        """
        unique = self.unique_four_vectors()
        lut = self.fourvec_to_indx(template, unique)
        ref_indices = [lut[FourVector(0, 0, 0, i)] for i in range(self.generator.num_sub_lattices)]

        template_cluster_list = deepcopy(self.clusters)
        for cluster in template_cluster_list:
            cluster.ref_indx = int(ref_indices[cluster.group])
            if cluster.size in {0, 1}:
                cluster.indices = []
            else:
                cluster.indices = self.generator.to_atom_index(cluster, lut)
        return template_cluster_list

    def unique_four_vectors(self) -> Set[FourVector]:
        """
        Return a list with all unique 4-vectors which are
        represented in any figure in all of the clusters.
        """
        # We utilize that FourVector objects are hashable,
        # and therefore can be filtered using a set()
        unique = set()
        for cluster in self.clusters:
            if cluster.size == 0:
                continue
            for figure in cluster.figures:
                for fv in figure.components:
                    unique.add(fv)
        return unique

    def get_figures(self) -> List[ase.Atoms]:
        """
        Return a list of atoms object representing the clusters
        """
        return self.clusters.get_figures(self.generator)

    def create_four_vector_lut(self, template: ase.Atoms) -> Dict[FourVector, int]:
        """
        Construct a lookup table (LUT) for the index in template given the
        wrapped vector

        Parameter:

        template: Atoms
            Atoms object to use when creating the lookup table (LUT)
        """
        lut = {}
        pos = template.get_positions().copy()
        for i in range(pos.shape[0]):
            if self.is_background_atom(template[i]):
                # No need to make a lookup for a background atom
                continue
            vec = self.generator.to_four_vector(pos[i, :], template[i].tag)
            lut[vec] = i
        return lut

    def fourvec_to_indx(
        self, template: ase.Atoms, unique: Sequence[FourVector]
    ) -> Dict[FourVector, int]:
        """Translate a set of unique FourVectors into their corresponding index
        in a template atoms object."""
        cell = template.get_cell()
        pos = np.zeros((len(unique), 3))
        for i, fv in enumerate(unique):
            pos[i, :] = fv.to_cartesian(self.prim)

        pos = wrap_positions(pos, cell)
        unique_indices = []
        for i in range(pos.shape[0]):
            diff_sq = np.sum((pos[i, :] - template.get_positions()) ** 2, axis=1)
            unique_indices.append(np.argmin(diff_sq))
        return dict(zip(unique, unique_indices))

    def translation_matrix(self, template: ase.Atoms) -> List[Dict[int, int]]:
        """
        Construct the translation matrix.

        The translation matrix translates a given atomic index to
        the corresponding atomic site if we started from index 0.

        Parameter:

        template: ase.Atoms
            Atoms object representing the simulation cell
        """
        cell = template.get_cell()

        # Get the unique four-vectors which are present in all of our clusters.
        unique = self.unique_four_vectors()

        # Only set "trivial_supercell" to True, if we allow that path.
        # Setting self._allow_trivial_path to False is only for testing purposes
        if self._allow_trivial_path:
            trivial_supercell = tools.is_trivial_supercell(self.prim, template)
        else:
            trivial_supercell = False

        # Set up the FourVector wrap function
        if trivial_supercell:
            nx, ny, nz = tools.get_repetition(self.prim, template)
            wrap_fnc = functools.partial(
                self._wrap_four_vectors_trivial, unique=unique, nx=nx, ny=ny, nz=nz
            )
            logger.info("Trivial supercell with repetition: (%d, %d, %d)", nx, ny, nz)
        else:
            # Choose the generalized pathway.
            wrap_fnc = functools.partial(self._wrap_four_vectors_general, unique=unique, cell=cell)
            logger.info("Non-trivial supercell, will wrap using cartesian coordinates")

        lut = self.create_four_vector_lut(template)

        # Map the un-translated unique 4-vectors to their index
        unique_indx_lut = self.fourvec_to_indx(template, unique)
        # call int() to convert from NumPy integer to python integer
        unique_index = [int(unique_indx_lut[u]) for u in unique]

        def _make_site_mapping(atom: ase.Atom) -> Dict[int, int]:
            """Helper function to calculate the translation mapping for each
            atomic site."""
            if self.is_background_atom(atom):
                # This atom is considered a background, it has no mapping
                return {}

            # Translate the atom into its four-vector representation
            vec = self.generator.to_four_vector(atom.position, sublattice=atom.tag)

            # Calculate the four vectors wrapped back into the supercell
            four_vecs = wrap_fnc(vec)

            # Get the index of the translated four-vector
            indices = [lut[fv] for fv in four_vecs]

            return dict(zip(unique_index, indices))

        # Calculate the mapping for each site in the template.
        trans_mat = list(map(_make_site_mapping, template))
        return TransMatrix(trans_mat)

    def _wrap_four_vectors_trivial(
        self,
        translation_vector: FourVector,
        unique: Sequence[FourVector],
        nx: int,
        ny: int,
        nz: int,
    ) -> Iterator[FourVector]:
        """Wrap FourVectors using the trivial shift+modulo operation"""
        # pylint: disable=no-self-use
        # Create as a generator, no need to assign this into a new list.
        return (u.shift_xyz_and_modulo(translation_vector, nx, ny, nz) for u in unique)

    def _wrap_four_vectors_general(
        self,
        translation_vector: FourVector,
        unique: Sequence[FourVector],
        cell: np.ndarray,
    ) -> Iterator[FourVector]:
        """Generalized FourVector wrapping function."""
        # Translate the (x, y, z) components of the unique four-vectors
        # by this atom's (x, y, z) four-vector component
        translated_unique = [u.shift_xyz(translation_vector) for u in unique]
        # Find the new Cartesian coordinates of the translated FourVectors,
        # and wrap them back into the cell
        cartesian = self.generator.to_cartesian(*translated_unique)
        cartesian = wrap_positions(cartesian, cell)

        # Re-translate the wrapped-Cartesian coordinates of the unique four-vectors
        # into a four-vector representation (with a generator expression)
        sublattices = [fv.sublattice for fv in translated_unique]
        return self.generator.many_to_four_vector(cartesian, sublattices)

    def build_all(
        self,
        template_atoms: ase.Atoms,
        max_cluster_dia: Sequence[float],
        index_by_sublattice: List[List[int]],
    ) -> Tuple[ClusterList, TransMatrix]:
        """Create a ClusterList and a TransMatrix object, and calculate the norm factors."""
        # Ensure that we have built the clusters for the cutoff
        self.build(max_cluster_dia)

        at_cpy = template_atoms.copy()
        # Build the cluster list and translation matrix.
        cluster_list = self.cluster_list_for_template(at_cpy)
        trans_matrix = self.translation_matrix(at_cpy)

        # Finally calculate the norm factors, and insert them into the new cluster list
        _set_norm_factors(
            index_by_sublattice,
            trans_matrix,
            cluster_list,
        )
        return cluster_list, trans_matrix


def _set_norm_factors(
    index_by_sublattice: List[List[int]],
    trans_matrix: TransMatrix,
    cluster_list: ClusterList,
) -> None:
    """Set normalization factor for each cluster. Will mutate the ClusterList.

    The normalization factor only kicks in when the cell is too small and
    thus, include self-interactions. This methods corrects the impact of
    self-interactions.
    """
    symm_group = np.zeros(trans_matrix.n_sites, dtype=np.uintc)
    for num, group in enumerate(index_by_sublattice):
        symm_group[group] = num

    for cluster in cluster_list:
        fig_keys = list(set(cluster.get_all_figure_keys()))
        num_occ = {}
        for key in fig_keys:
            num_occ[key] = cluster_list.num_occ_figure(key, cluster.name, symm_group, trans_matrix)
        num_fig_occ = cluster.num_fig_occurences
        norm_factors = {}
        for key in fig_keys:
            tot_num = num_occ[key]
            num_unique = len(set(key.split("-")))
            norm_factors[key] = float(tot_num) / (num_unique * num_fig_occ[key])

        norm_factor_list = []
        for fig in cluster.indices:
            key = cluster.get_figure_key(fig)
            norm_factor_list.append(norm_factors[key])
        assert len(norm_factor_list) == len(cluster.indices)
        cluster.info["normalization_factor"] = norm_factor_list


class _CacheChecker:
    """Helper class to check if the
    cluster manager has already been built
    using a given set of settings.
    """

    def __init__(self):
        self.max_cluster_dia = None

    def requires_build(self, max_cluster_dia: Sequence[float]) -> bool:
        """Check if a given set of 'max_size'
        and 'max_cluster_dia' has previously been
        used to build the clusters"""
        # We don't have anything cached yet
        if self.max_cluster_dia is None:
            return True

        # Check if the parameters match
        return not np.array_equal(self.max_cluster_dia, max_cluster_dia)

    def set_cache(self, max_cluster_dia: Sequence[float]):
        # Ensure we set a copy, so no external
        # mutations affect the cache
        self.max_cluster_dia = deepcopy(max_cluster_dia)



from typing import Union
from abc import ABC
from ase import Atoms
from .mc_evaluator import MCEvaluator, construct_evaluator

__all__ = ("BaseMC",)


class BaseMC(ABC):
    """Base Monte Carlo Class.
    Initializes the internal atoms and evaluator objects.

    Args:
        system (Union[Atoms, MCEvaluator]): Either an ASE Atoms object
            with an attached calculator, or a pre-initialized
            :class:`~clease.montecarlo.mc_evaluator.MCEvaluator`
            object.
        temp (float): Temperature of Monte Carlo simulation in Kelvin
    """

    def __init__(self, system: Union[Atoms, MCEvaluator], temp: float):
        self.evaluator = system
        self.temperature = temp

    @property
    def evaluator(self) -> MCEvaluator:
        """The internal evaluator object.

        :getter: Returns the internal
            :class:`~clease.montecarlo.mc_evaluator.MCEvaluator` object.
        :setter: Sets the internal evaluator object. Can either accept an
            atoms object, or a pre-initialized evaluator object.
            See ``system`` in the docstring of the class constructor.
        :type: :class:`~clease.montecarlo.mc_evaluator.MCEvaluator`
        """
        return self._evaluator

    @evaluator.setter
    def evaluator(self, value: Union[Atoms, MCEvaluator]) -> None:
        """Set the evaluator object. If the value is an Atoms object,
        the evaluator is created on basis of the attached calculator object.
        """
        self._evaluator = construct_evaluator(value)

    @property
    def atoms(self) -> Atoms:
        """The internal Atoms object.

        :type: :class:`ase.atoms.Atoms`
        """
        return self.evaluator.atoms

    def _get_temperature(self) -> float:
        """Retrieve the temperature of the MC."""
        return self._temperature

    def _set_temperature(self, value: float) -> None:
        """Set the internal temperature."""
        self._temperature = value  # pylint: disable=attribute-defined-outside-init
        self._on_temp_change()

    def _on_temp_change(self) -> None:
        """An MC object may choose to make adjustments after a temperature change,
        by subclassing and adjusting this method."""

    # Ensure that changing the temperature triggers the correct responses.
    temperature = property(
        _get_temperature,
        _set_temperature,
        doc="""
        Property for getting and setting the temperature of the MC object.

        :type: float
        """,
    )
    # For backwards compatibility, alias the "temperature" as "T" as well
    T = property(
        _get_temperature,
        _set_temperature,
        doc="""
                 Alias for the temperature variable.
                 This variable name is deprecated in favor of :py:attr:`~temperature`.

                 :type: float
                 """,
    )



from __future__ import annotations
from typing import List, Dict
import attr
import numpy as np

# Constant used for indicating a missing value in the NumPy array.
# Indicates a background index.
_MISSING = -1

__all__ = ("TransMatrix",)


@attr.define(eq=True, order=False)
class TransMatrix:
    """Data class wrapper for the translation matrix."""

    trans_matrix: List[Dict[int, int]] = attr.field()

    def key_array(self) -> np.ndarray:
        """The keys of the trans matrix in NumPy format."""
        return self.as_array()[:, :, 0]

    def value_array(self) -> np.ndarray:
        """The values of the trans matrix in NumPy format."""
        return self.as_array()[:, :, 1]

    def as_array(self) -> np.ndarray:
        """Convert the trans matrix into its NumPy array format.
        "Missing" values are designated the value of -1.

        This representation is more convenient for storage.
        """
        arr = np.zeros((len(self), self.n_indices, 2), dtype=np.dtype(int))
        for ii, dct in enumerate(self.trans_matrix):
            if len(dct) == 0:
                # Populate with -1's, indiate missing, i.e. background atoms
                arr[ii, :, :] = _MISSING
                continue
            arr[ii, :, 0] = list(dct.keys())
            arr[ii, :, 1] = list(dct.values())
        return arr

    @property
    def n_indices(self) -> int:
        """Number of index pairs for indices which are not background."""
        # Some entries may have 0 length, if they are background atoms
        # So we have to check the lengths. The validation already checked
        # we only have at most 2 possible values.
        try:
            return max(len(dct) for dct in self.trans_matrix)
        except ValueError:
            # Trans matrix has length 0
            return 0

    @property
    def n_sites(self) -> int:
        """The number of sites which were used to build the trans matrix."""
        return len(self)

    @trans_matrix.validator
    def _validate_trans_matrix(self, attribute, value):
        # pylint: disable=unused-argument, no-self-use
        if not isinstance(value, list):
            raise TypeError(f"Expected the trans matrix as a list, got {value!r}")
        lens = set(len(dct) for dct in value)
        if len(lens) not in {0, 1, 2}:
            # length 0: No entries in the TM
            # length 1: No background atoms, every entry is filled
            # length 2: Background atoms has a length 0, so we should have 0's
            #  and then whatever number of pairs
            raise ValueError(f"Got too many different lengths, expected 0, 1 or 2, got {len(lens)}")

    @classmethod
    def from_array(cls, array: np.ndarray) -> TransMatrix:
        """Construct the TransMatrix from it's NumPy array representation,
        as exported from the "as_array" method.
        """
        # Sanity checks
        # We sometimes load this object from a h5py dataset, which is just a wrapper around the
        # numpy array, so we want to avoid making a copy into an entirely new NumPy array,
        # since there's no need.
        if not isinstance(array, np.ndarray):
            raise ValueError(f"Expected a NumPy array, got {array!r}")
        if array.ndim != 3:
            raise ValueError(f"Wrong number of dimensions, expected 3, got {array.ndim}")
        if array.shape[-1] != 2:
            raise ValueError(f"Last dimension should be 2, got {array.shape[-1]}")

        def _convert_array(arr: np.ndarray) -> dict:
            """Helper function to convert the array."""
            if (arr == _MISSING).any():
                # This is a background atom, every entry should be "MISSING"
                assert (arr == _MISSING).all()
                return {}
            # "zip" the last dimension together as key-value pairs into a dict
            return dict(arr)

        tm = list(map(_convert_array, array))
        return cls(tm)

    def __len__(self) -> int:
        return len(self.trans_matrix)

    def __getitem__(self, i):
        return self.trans_matrix[i]



from .settings import ClusterExpansionSettings

__all__ = ("settings_from_json",)


def settings_from_json(fname) -> ClusterExpansionSettings:
    """Initialize settings from JSON.

    Exists due to compatibility. You should instead use
    `ClusterExpansionSettings.load(fname)`

    Parameters:

    fname: str
        JSON file where settings are stored
    """
    return ClusterExpansionSettings.load(fname)  # pylint: disable=no-member



# pylint: disable=undefined-variable
# For some reason, pylint detects a lot of cyclic imports here.
from .concentration import *
from . import template_filters
from . import template_atoms
from . import atoms_manager
from .settings import *
from .settings_bulk import *
from .settings_slab import *
from .utils import *

from . import settings

ADDITIONAL = ("settings", "template_atoms", "template_filters", "atoms_manager")
__all__ = (
    concentration.__all__
    + settings.__all__
    + settings_bulk.__all__
    + settings_slab.__all__
    + utils.__all__
    + ADDITIONAL
)



from typing import Sequence, List, Dict
import random
import numpy as np
import ase


class SwapMoveIndexTracker:
    """
    Tracker for possible swap move indices.

    :param atoms: The ase Atoms object to be tracked.
    :param indices: List with indices that can be chosen from. If None, all indices
        can be chosen.
    """

    def __init__(self, atoms, indices: Sequence[int] = None):
        self.tracker: Dict[str, List[int]] = {}
        self.index_loc = None
        self._last_move = []

        self._init_tracker(atoms, indices=indices)

        # Cache of the unique symbols for faster lookup
        self.unique_symbols = self.get_unique_symbols()

    @staticmethod
    def _unique_symbols_from_atoms(atoms) -> List[str]:
        return sorted(set(atoms.symbols))

    def get_unique_symbols(self) -> List[str]:
        return list(self.tracker.keys())

    @property
    def num_symbols(self) -> int:
        return len(self.tracker)

    def __repr__(self):
        str_repr = f"SwapMoveIndexTracker at {hex(id(self))}\n"
        str_repr += f"Symbols tracked: {self.get_unique_symbols()}\n"
        str_repr += f"Tracker info: {self.tracker}\n"
        return str_repr

    def _init_tracker(self, atoms: ase.Atoms, indices: Sequence[int] = None) -> None:
        """Initialize the tracker with the numbers."""
        # Cache the unique symbols for faster access
        symbols = self._unique_symbols_from_atoms(atoms)

        # Track indices of all symbols
        self.tracker = {symb: [] for symb in symbols}

        # Track the location in self.tracker of each index
        self.index_loc = np.zeros(len(atoms), dtype=int)
        for atom in atoms:
            self.tracker[atom.symbol].append(atom.index)
            self.index_loc[atom.index] = len(self.tracker[atom.symbol]) - 1

        if indices is not None:
            self._filter_indices(indices)

    def _filter_indices(self, include: Sequence[int]) -> None:
        include_set = set(include)
        for s, idx in self.tracker.items():
            self.tracker[s] = list(set(idx).intersection(include_set))
            for i, num in enumerate(self.tracker[s]):
                self.index_loc[num] = i

        # Remove empty items
        self.tracker = {k: v for k, v in self.tracker.items() if v}

    def move_already_updated(self, system_changes) -> bool:
        """Return True if system_changes have already been taken into account.

        Parameters:

        system_changes:
        """
        return system_changes == self._last_move

    def update_swap_move(self, system_changes) -> None:
        """Update the atoms tracker.

        Parameters:

        system_changes:
        """
        if self.move_already_updated(system_changes):
            # This change has already been updated!
            return
        self._last_move = system_changes
        indx1 = system_changes[0].index
        indx2 = system_changes[1].index
        symb1 = system_changes[0].old_symb
        symb2 = system_changes[0].new_symb

        # Find the locations of the indices
        loc1 = self.index_loc[indx1]
        loc2 = self.index_loc[indx2]

        # Update the tracker and the locations
        self.tracker[symb1][loc1] = indx2
        self.index_loc[indx2] = loc1

        self.tracker[symb2][loc2] = indx1
        self.index_loc[indx1] = loc2

    def undo_last_swap_move(self) -> None:
        """Undo last swap move."""
        if not self._last_move:
            return
        opposite_change = []
        for change in self._last_move:
            opposite_change.append((change[0], change[2], change[1]))

        self.update_swap_move(opposite_change)
        self._last_move = []

    def get_random_indx_of_symbol(self, symbol: str) -> int:
        return random.choice(self.tracker[symbol])



import click
from clease.db_util import get_all_cf_names, get_all_cf, get_cf_tables
from . import main


@main.clease_cli.group()
def db():
    """The DB CLI"""


@db.command(help="Display the available correlation function tables")
@click.argument("db_name", type=str)
def tab(db_name):
    show_cf_tables(db_name)


@db.command(help="Display the available CF names in the DB")
@click.argument("db_name", type=str)
def names(db_name):
    show_cf_names(db_name)


@db.command(help="Display the correlation functions for a specific ID in the database")
@click.argument("db_name", type=str)
@click.argument("db_id", type=int)
def cf(db_name, db_id):
    try:
        show_cf(db_name, db_id)
    except Exception as exc:  # pylint: disable=broad-except
        click.echo(f"An error occurred: {exc}")


def show_cf_tables(db_name: str):
    tables = get_cf_tables(db_name)
    cftab = ", ".join(tables)
    click.echo(f"Available correlation function tables: {cftab}")


def show_cf_names(db_name: str):
    table = get_cf_tables(db_name)[0]
    cf_names = sorted(get_all_cf_names(db_name, table))
    click.echo("Tracked correlation functions")

    line = ""
    char_per_line = 110
    for name in cf_names:
        if len(line) + len(name) > char_per_line:
            click.echo(line)
            line = ""
        if line == "":
            line += name
        else:
            line += ", " + name
    click.echo(line)


def show_cf(db_name: str, db_id: int):
    table = get_cf_tables(db_name)[0]
    cfs = get_all_cf(db_name, table, db_id)
    click.echo(f"Correlation functions for ID {db_id}")
    for k, v in cfs.items():
        click.echo(f"{k}: {v}")



# distutils: language = c++
# cython: c_string_type=str, c_string_encoding=ascii
import numpy as np
cimport numpy as np  # Initialize the Numpy API
np.import_array()
include "pyce_updater.pyx"
include "py_cluster.pyx"
include "py_atoms.pyx"

# Files that use the Numpy Array API are included here
cdef extern from "with_numpy/ce_updater.cpp":
    pass



from typing import List
from abc import ABCMeta
from ase import Atoms
from ase.neighborlist import neighbor_list

__all__ = ("KMCEventType", "NeighbourSwap")


class KMCEventType(metaclass=ABCMeta):
    @property
    def name(self):
        return type(self).__name__

    def get_swaps(self, atoms: Atoms, vac_idx: int) -> List[int]:
        raise NotImplementedError(f"get_swaps has not been implemented for class {self.name}")


class NeighbourSwap(KMCEventType):
    """
    KMC event that provides swaps between neighbours

    :param cutoff: Passed to ASE neighbour list. All sites
        within the cutoff are possible swaps
    """

    def __init__(self, atoms: Atoms, cutoff: float):
        super().__init__()
        first, second = neighbor_list("ij", atoms, cutoff)
        self.nl = [[] for _ in range(len(atoms))]
        for f, s in zip(first, second):
            self.nl[f].append(s)

    def get_swaps(self, atoms: Atoms, vac_idx: int) -> List[int]:
        return self.nl[vac_idx]



from clease.datastructures.system_changes import SystemChanges
from .mc_constraint import MCConstraint


class FixedElement(MCConstraint):
    """
    Class for forcing an element of a certiain type to stay fixed.

    Parameters:

    element: str
        Name of the element that is supposed to stay fixed
    """

    def __init__(self, element):
        self.element = element

    def __call__(self, system_changes: SystemChanges):
        for change in system_changes:
            if self.element in (change.old_symb, change.new_symb):
                return False
        return True



"""Module for tools pertaining to geometry of atoms and cells."""
import itertools
import numpy as np

__all__ = ("max_sphere_dia_in_cell",)


def max_sphere_dia_in_cell(cell: np.ndarray) -> float:
    """Find the diameter of the largest possible sphere which can be placed
    inside a cell.

    For example, how large of a Death Star could be built inside a given atoms object?

    Args:
        cell (np.ndarray): A (3 x 3) matrix which defines the cell parameters.
            Raises a ValueError if the cell shape is wrong.

    Returns:
        float: The diameter of the largest sphere which can fit within the cell.
    """
    if not cell.shape == (3, 3):
        raise ValueError(f"Cell should be (3, 3), got {cell.shape}")
    # Midpoint of the parallelogram spanned by the cell vectors
    midpoint = np.sum(cell / 2, axis=0)

    # Get the shortest distance between the midpoint and planes defined by the
    # ab, ac and bc vectors. Since the cell is a parallelogram, the 3 remaining planes will
    # have identical distances to the midpoint.
    # Assumption: One of the cell corners is at (0, 0, 0),
    # so 3 of the 6 planes will share the point (0, 0, 0).
    # PQ is the vector from our midpoint to the reference point in the plane.
    PQ = midpoint - [0, 0, 0]

    def _iter_distances() -> float:
        # Iterate all distances from cell wall to the midpoint
        for v1, v2 in itertools.combinations(cell, 2):
            # Find the normal vector to the plane spanned by v1 and v2
            # Note: The astype(float) is to ensure that a cell defined by
            # integers still works.
            n = np.cross(v1, v2).astype(float)
            n /= np.linalg.norm(n)
            # Project onto the normal vector to get the shortest possible
            # distance.
            # Multiply by 2 to get the diameter.
            yield np.linalg.norm(PQ.dot(n)) * 2

    # The smallest possible distance is the radius of the largest sphere we can have.
    return min(_iter_distances())



"""Collection of classess to perform regression."""
from typing import Union, Optional, List
import numpy as np
from numpy.linalg import pinv
from sklearn.linear_model import Lasso as skLasso
from clease.data_normalizer import DataNormalizer

__all__ = ("LinearRegression", "Tikhonov", "Lasso")


class LinearRegression:
    def __init__(self) -> None:
        self._weight_matrix = None
        self.tol = 1e-8

    @property
    def weight_matrix(self):
        return self._weight_matrix

    def _ensure_weight_matrix_consistency(self, data: np.ndarray) -> None:
        """Raise an error if the dimensions of the
           weight matrix is not consistent.

        :param data: y-values in the fit
        """
        if self._weight_matrix is not None:
            if self._weight_matrix.shape[1] != len(data):
                raise ValueError(
                    f"The provided weight matrix needs to have "
                    f"dimension {len(data)}x{len(data)}, "
                    f"{self._weight_matrix.shape} given"
                )

    def fit(self, X: np.ndarray, y: np.ndarray) -> np.ndarray:
        """Fit a linear model by performing ordinary least squares

        y = Xc

        :param X: Design matrix (NxM)

        :param y: Data points (vector of length N)
        """
        self._ensure_weight_matrix_consistency(y)

        # We use SVD to carry out the fit
        U, D, V_h = np.linalg.svd(X, full_matrices=False)
        V = V_h.T
        diag_item = np.zeros_like(D)
        mask = np.abs(D) > self.tol
        diag_item[mask] = 1.0 / D[mask]
        coeff = V.dot(np.diag(diag_item)).dot(U.T).dot(y)
        return coeff

    def precision_matrix(self, X: np.ndarray) -> np.ndarray:
        D, V_h = np.linalg.svd(X, full_matrices=False)[1:]  # U is unused
        V = V_h.T
        diag = np.zeros_like(D)
        mask = np.abs(D) > self.tol
        diag[mask] = 1.0 / D[mask] ** 2
        return V.dot(np.diag(diag)).dot(V.T)

    def get_instance_array(self) -> List[object]:
        # pylint: disable=no-self-use
        return [LinearRegression()]

    def is_scalar(self):
        # pylint: disable=no-self-use
        return False

    def get_scalar_parameter(self):  # pragma: no cover
        # pylint: disable=no-self-use
        raise ValueError("Fitting scheme is not described by a scalar parameter!")

    @property
    def support_fast_loocv(self):
        return True


class Tikhonov(LinearRegression):
    """Ridge regularization.

    :param alpha: regularization term

        - float: A single regularization coefficient is used for all features.
                 Tikhonov matrix is T = alpha * I (I = identity matrix).

        - 1D array: Regularization coefficient is defined for each feature.
                    Tikhonov matrix is T = diag(alpha) (the alpha values are
                    put on the diagonal).
                    The length of array should match the number of features.

        - 2D array: Full Tikhonov matrix supplied by a user.
                    The dimensions of the matrix should be M * M where M is the
                    number of features.

    :param normalize: If True each feature will be normalized to before fitting
    """

    def __init__(
        self,
        alpha: Union[float, np.ndarray] = 1e-5,
        penalize_bias_term: bool = False,
        normalize: bool = True,
    ) -> None:
        super().__init__()
        self.alpha = alpha
        self.penalize_bias_term = penalize_bias_term
        self.normalize = normalize

    def _get_tikhonov_matrix(self, num_clusters: int) -> np.ndarray:
        if isinstance(self.alpha, np.ndarray):
            if len(self.alpha.shape) == 1:
                tikhonov = np.diag(self.alpha)
            elif len(self.alpha.shape) == 2:
                tikhonov = self.alpha
            else:
                raise ValueError("Matrix have to have dimension 1 or 2")
        else:
            # Alpha is a floating point number
            tikhonov = np.identity(num_clusters)
            if not self.penalize_bias_term:
                tikhonov[0, 0] = 0.0
            tikhonov *= np.sqrt(self.alpha)
        return tikhonov

    def fit(self, X: np.ndarray, y: np.ndarray) -> np.ndarray:
        """Fit coefficients based on Ridge regularizeation."""
        self._ensure_weight_matrix_consistency(y)

        if self.weight_matrix is None:
            W = np.ones(len(y))
        else:
            W = np.diag(self.weight_matrix)

        X_fit = X
        y_fit = y
        if self.normalize:
            if np.any(np.abs(X[:, 0] - 1.0) > 1e-16):
                msg = "Tikhonov: When normalize=True. The first column in X "
                msg += "should correspond to a to a bias term. Therefore, all "
                msg += f"entries should be 1.\nGot:\n{X[:, 0]}\n"
                raise ValueError(msg)
            normalizer = DataNormalizer()
            X_fit, y_fit = normalizer.normalize(X[:, 1:], y)

        precision = self.precision_matrix(X_fit)
        coeff = precision.dot(X_fit.T.dot(W * y_fit))

        if self.normalize:
            coeff_with_bias = np.zeros(len(coeff) + 1)
            coeff_with_bias[1:] = normalizer.convert(coeff)
            coeff_with_bias[0] = normalizer.bias(coeff)
            coeff = coeff_with_bias
        return coeff

    def precision_matrix(self, X: np.ndarray) -> np.ndarray:
        """Calculate the presicion matrix."""
        num_features = X.shape[1]
        tikhonov = self._get_tikhonov_matrix(num_features)

        if tikhonov.shape != (num_features, num_features):
            raise ValueError(
                "The dimensions of Tikhonov matrix do not match " "the number of clusters!"
            )

        W = self.weight_matrix
        if W is None:
            W = np.eye(X.shape[0])
        precision = pinv(X.T.dot(W.dot(X)) + tikhonov.T.dot(tikhonov))
        return precision

    @staticmethod
    def get_instance_array(
        alpha_min: float,
        alpha_max: float,
        num_alpha: int = 10,
        scale: Optional[str] = "log",
    ) -> List[object]:
        # pylint: disable=arguments-differ
        if scale == "log":
            alpha = np.logspace(
                np.log10(alpha_min), np.log10(alpha_max), int(num_alpha), endpoint=True
            )
        else:
            alpha = np.linspace(alpha_min, alpha_max, int(num_alpha), endpoint=True)
        return [Tikhonov(alpha=a) for a in alpha]

    def is_scalar(self) -> bool:
        return isinstance(self.alpha, float)

    def get_scalar_parameter(self) -> float:
        if self.is_scalar():
            return self.alpha
        return super().get_scalar_parameter()


class Lasso(LinearRegression):
    """LASSO regularization.

    :param alpha: regularization coefficient
    :param max_iter: (int) Maximum number of iterations.
    """

    def __init__(self, alpha: float = 1e-5, max_iter: int = int(1e6)) -> None:
        super().__init__()
        self.alpha = alpha
        self.max_iter = max_iter

    @property
    def max_iter(self) -> int:
        return self._max_iter

    @max_iter.setter
    def max_iter(self, value) -> None:
        # sklearn Lasso class must have integer max_iter values.
        self._max_iter = int(value)

    def fit(self, X: np.ndarray, y: np.ndarray) -> np.ndarray:
        """Fit coefficients based on LASSO regularizeation."""
        lasso = skLasso(alpha=self.alpha, fit_intercept=False, copy_X=True, max_iter=self.max_iter)
        lasso.fit(X, y)
        return lasso.coef_

    @property
    def weight_matrix(self):
        return LinearRegression.weight_matrix(self)

    @staticmethod
    def get_instance_array(
        alpha_min: float, alpha_max: float, num_alpha: int = 10, scale: str = "log"
    ) -> List[object]:
        # pylint: disable=arguments-differ
        if scale == "log":
            alpha = np.logspace(
                np.log10(alpha_min), np.log10(alpha_max), int(num_alpha), endpoint=True
            )
        else:
            alpha = np.linspace(alpha_min, alpha_max, int(num_alpha), endpoint=True)
        return [Lasso(alpha=a) for a in alpha]

    def is_scalar(self):
        return True

    def get_scalar_parameter(self) -> float:
        return self.alpha

    def precision_matrix(self, X):  # pragma: no cover
        raise NotImplementedError("Precision matrix for LASSO is not implemented.")

    @property
    def support_fast_loocv(self):
        return False



"""Import the C++ definition of a cluster to Cython"""
# distutils: language = c++

from libcpp.vector cimport vector

cdef extern from "cluster.hpp":
    cdef cppclass Cluster:
        Cluster(object info_dict) except +

        vector[vector[int]] get_all_decoration_numbers(int n_basis_funcs) const

        int get_size() const



from typing import List, Union
from pathlib import Path
import numpy as np
from .buffered_array import BufferedArray

__all__ = ("EntropyProductionRate",)


class EntropyProductionRate:
    """
    Tracks entropy production rate (EPR) using a Gallavotti-Cohen functional.

    EPR = 1/N sum_{i=0}^N ln P(i -> j)/P(j -> i)

    N is the number of steps of a path and P(i -> j) is the probability of
    going from state i to state j. The expression is exact in the limit
    N -> infty. However, this class tracks the terms inside the sum and write
    them to file. To calculate the time evolution of EPR one can use a windowed
    average of the resulting data.

    References:

    [1] Gourgoulias, Konstantinos, Markos A. Katsoulakis, and Luc Rey-Bellet.
        "Information criteria for quantifying loss of reversibility in parallelized KMC."
        Journal of Computational Physics 328 (2017): 438-454.

    :param buffer_length: Length of buffer used to temporarily store the terms
        in the sum in memory. When the buffer is full, it is flushed to a text
        file.
    :param logfile: Filename of the file used when the buffer is flushed
    """

    def __init__(self, buffer_length: int = 10000, logfile: Union[str, Path] = "epr.txt"):
        self._buffer = BufferedArray(size=buffer_length, fname=logfile)
        self.prev_swap = -1
        self.prob_forw = 0.0

    def reset(self):
        """
        Clear all information stored
        """
        self.prev_swap = -1
        self.prob_forw = 0.0
        self._buffer.clear()

    @staticmethod
    def _probabilty(choice: int, cumulative_rates: np.ndarray):
        """
        Calculate the probability of going from the current state to the
        chosen state.
        """
        if choice == 0:
            return cumulative_rates[0]
        return cumulative_rates[choice] - cumulative_rates[choice - 1]

    def update(self, current: int, choice: int, cumulative_rates: np.ndarray, swaps: List[int]):
        """
        Update the buffer

        :param current: Current position of the vacancy
        :param choice: Index into cumulative_rates that is chosen
        :param cumulative_rates: Cumulative sum of the rates
        :param swaps: Possible swaps
        """
        if self.prev_swap == -1:
            # First, time we can't calculate the backward rate.
            # Track the swap and return
            self.prev_swap = current
            self.prob_forw = self._probabilty(choice, cumulative_rates)
            return

        backward = swaps.index(self.prev_swap)
        prob_back = self._probabilty(backward, cumulative_rates)
        self.prev_swap = current
        self._buffer.push(np.log(self.prob_forw / prob_back))

    def flush_buffer(self):
        self._buffer.flush()



# pylint: disable=undefined-variable
from . import swap_move_index_tracker
from .mc_evaluator import *
from .barrier_models import *
from .kmc_events import *
from .kinetic_monte_carlo import *
from .trial_move_generator import *
from . import base
from .bias_potential import BiasPotential
from .montecarlo import Montecarlo
from .sgc_montecarlo import SGCMonteCarlo
from .gaussian_kernel_bias_potential import GaussianKernelBiasPotential
from .binned_bias_potential import BinnedBiasPotential
from .random_bias_with_memory import RandomBiasWithMemory
from .metadynamics_sampler import MetaDynamicsSampler

ADDITIONAL = (
    "Montecarlo",
    "SGCMonteCarlo",
    "MetaDynamicsSampler",
    "BinnedBiasPotential",
    "BiasPotential",
    "GaussianKernelBiasPotential",
    "RandomBiasWithMemory",
    "base",
    "swap_move_index_tracker",
)
__all__ = ADDITIONAL + (
    barrier_models.__all__
    + kinetic_monte_carlo.__all__
    + kmc_events.__all__
    + trial_move_generator.__all__
    + mc_evaluator.__all__
)



class Averager:
    """
    Class for robust averaging.

    Parameter:

    ref_value: float
        Reference value used to normalise data internally.
    """

    def __init__(self, ref_value=1.0):
        self._ref_value = float(ref_value)

        if abs(self._ref_value) < 1e-6:
            self._ref_value = 1.0
        self._n_samples = 0
        self._mean = 0.0

    def __iadd__(self, value):
        """
        += Operator

        Parameter:

        value: float
            Value to be added
        """
        if isinstance(value, Averager):
            self._mean += value._mean * (value._ref_value / self._ref_value)
            self._n_samples += value._n_samples
            return self

        self._n_samples += 1
        self._mean += value / self._ref_value
        return self

    def __add__(self, other):
        """Add to Averager objects.

        Parameter:

        other: Averager
            Another Averager object to be added
        """
        if not isinstance(other, Averager):
            raise NotImplementedError

        ratio = other._ref_value / self._ref_value
        new_obj = Averager(ref_value=self._ref_value)
        new_obj._mean = self._mean + other._mean * ratio
        new_obj._n_samples = self._n_samples + other._n_samples
        return new_obj

    def __itruediv__(self, number):
        return self.__idiv__(number)

    def __idiv__(self, number):
        """Divide by number.

        Parameter:

        number: float
            Number to divide by
        """
        self._mean /= float(number)
        self._n_samples /= float(number)
        return self

    def clear(self):
        """
        Clear the accumulated entries, but keep the reference value
        """
        self._n_samples = 0
        self._mean = 0.0

    @property
    def mean(self):
        if self._n_samples == 0:
            return self._mean * self._ref_value
        return (self._mean / self._n_samples) * self._ref_value



from deprecated import deprecated
from . import settings

__all__ = ("Concentration",)


@deprecated(version="0.10.0", reason="Use Concentration class from clease.settings instead")
class Concentration(settings.Concentration):
    """Concentration class was moved to clease.settings"""



"""Import the C++ definition of a FourVector to Cython"""
# distutils: language = c++

cdef extern from "four_vector.hpp":
    cdef cppclass FourVector:
        FourVector()

        int ix, iy, iz, sublattice



import os
from pathlib import Path
import sysconfig
from setuptools import setup, Extension


def src_folder():
    candidates = ["cxx/src/", "clease/", "./"]
    for c in candidates:
        if os.path.exists(os.path.join(c, "cluster.cpp")):
            return c
    raise RuntimeError("Cannot find source folder.")


def get_cython_folder():
    candidates = ["cxx/cython/", "cython/", "./"]
    for c in candidates:
        if os.path.exists(os.path.join(c, "clease_cxx.pyx")):
            return c
    raise RuntimeError("Cannot find cython folder")


def include_folder():
    candidates = ["cxx/include/", "clease/", "./"]
    for c in candidates:
        if os.path.exists(os.path.join(c, "cluster.hpp")):
            return c
    raise RuntimeError("Cannot find include folder")


def get_npy_include_folder():
    import numpy as np

    return np.get_include()


def build_ext(ext_module):
    from Cython.Build import cythonize

    return cythonize(
        ext_module, compiler_directives={"language_level": "3", "embedsignature": True}
    )


cxx_src_folder = src_folder()
cxx_inc_folder = include_folder()
cython_folder = get_cython_folder()


def get_source_files():
    """Construct a list of all sources to be included."""
    cpp_sources = Path(cxx_src_folder).glob("*.cpp")

    # C++ sources, excluding the ones which rely on NumPy
    # those are included directly from the Cython source
    sources = [str(f) for f in cpp_sources]

    # Cython sources
    cython_src = ["clease_cxx.pyx"]
    for src in cython_src:
        sources.append(os.path.join(cython_folder, src))
    return sources


src_files = get_source_files()
extra_comp_args = ["-std=c++11"]
# Replaces the deprecated "distutils.sysconfig.get_python_inc()"
py_include = sysconfig.get_path("include")

clease_cxx = Extension(
    "clease_cxx",
    sources=src_files,
    include_dirs=[
        cxx_inc_folder,
        get_npy_include_folder(),
        cxx_src_folder,
        py_include,
        cython_folder,
    ],
    extra_compile_args=extra_comp_args,
    language="c++",
)

# Extra packages required with certain aspects of CLEASE, which are not required
# to run the base CLEASE package.
EXTRAS_REQUIRE = {
    # For building the documentation
    "doc": ("sphinx", "sphinx_rtd_theme"),
    # For running the CLEASE test suite
    "test": (
        "pytest",
        "pytest-mock",
        "mock",
        # Get the histrogram extras, for making nice histogram plots
        # with pytest-benchmark
        "pytest-benchmark[histogram]>=3.4.1",
        "tox>=3.24.0",
    ),
    # Extra nice-to-haves when developing CLEASE
    "dev": (
        "pip",
        "cython",
        "pre-commit",
        "ipython",
        "twine",
        "black>=22.1.0",  # Style formatting
        "pylint",
        "pyclean>=2.0.0",  # For removing __pycache__ and .pyc files
        "pytest-cov",
        "build",
    ),
    "gui": ("clease-gui",),
}
# Make an entry which installs all of the above in one go
EXTRAS_REQUIRE["all"] = tuple({package for key, tup in EXTRAS_REQUIRE.items() for package in tup})

setup(
    ext_modules=build_ext(clease_cxx),
    extras_require=EXTRAS_REQUIRE,
)



from typing import Tuple
import numpy as np
from .regression import LinearRegression

__all__ = ("ConstrainedRidge",)


class ConstrainedRidge(LinearRegression):
    """
    Tikhonov regularized fitting scheme that supports linear
    constraints.

    :param alpha: Vector or matrix with the regularization
        parameters. If it is a vector, it is assumed that it
        corresponds to the diagonal of the Thikonov matrix.
        The length of the vector must be the same as the
        number of features to be fitted.
    """

    def __init__(self, alpha: np.ndarray):
        super().__init__()
        self.alpha = alpha
        if len(alpha.shape) == 1:
            self.alpha = np.diag(alpha)
        self._constraint = None

    def add_constraint(self, A: np.ndarray, c: np.ndarray) -> None:
        """
        Add a constraint to the system of equations A.dot(coeff) = c

        :param A: Matrix describing the linear system of equation
        :param c: Right hand side of the linear system of equations
        """
        self._constraint = {
            "A": A,
            "c": c,
        }

    def kkt_system(self, X: np.ndarray, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Return the Karush-Kuhn-Tucker (KKT) system of equations. By solving
        the KKT system, one solves the problem

        minimize (X.dot(coeff) - y)^2 subject to A.dot(coeff) = c

        where A and c are set via `add_constraint`.

        The linear system of equations to be solved is given
        in block matrix form as

        | 2X^T.dot(X) + 2*alpha      A^T || coeff |   | 2X^T.dot(y) |
        |                                ||       | = |             |
        |         A                  0   ||  lamb |   |      c      |

        where coeff is the sought solution, and lamb is a set of Lagrange
        multipliers.
        """
        if self._constraint is None:
            return X.T.dot(X) + self.alpha, X.T.dot(y)

        A = self._constraint["A"]
        zero = np.zeros((A.shape[0], A.shape[0]))
        matrix = np.block([[2.0 * X.T.dot(X) + self.alpha, A.T], [A, zero]])
        rhs = 2.0 * X.T.dot(y)
        rhs_constraint = self._constraint["c"]
        return matrix, np.concatenate((rhs, rhs_constraint))

    def fit(self, X: np.ndarray, y: np.ndarray) -> np.ndarray:
        """
        Find the least square solution to the X*coeff = y,
        where the coefficient satisfies eventual constraints
        attached. If no constraints are added, this is equivalent
        to normal Tikhonov regularized fit. In this case, it is
        recommended to use the Tikhonov scheme instead.

        :param X: Design matrix
        :param y: target values
        """
        matrix, rhs = self.kkt_system(X, y)
        coeff = np.linalg.solve(matrix, rhs)
        return coeff[: X.shape[1]]



import json
from typing import Any, Dict
import attr
from ase.io import jsonio as aseio
from ase.utils import reader, writer

__all__ = ("encode", "decode", "read_json", "write_json", "jsonable")


def clease_default(obj: Any) -> Any:
    """Function for identifying CLEASE object types
    in the Encoder"""
    if hasattr(obj, "clease_objtype"):
        d = obj.todict()
        d["__clease_objtype__"] = obj.clease_objtype
        return d
    raise TypeError("Not a CLEASE object")


class CleaseEncoder(json.JSONEncoder):
    """Wrapper around jsonio encoder, for encoding clease object types.
    Includes the ase JSON encoder as well, for simple encoding of
    ASE and NumPy objects.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.encoders = [clease_default, aseio.MyEncoder().default]

    # Arguments differ, because they call the variable "o" instead of "obj"
    # in JSONEncoder. It doesn't look as nice.
    def default(self, obj):
        # pylint: disable=arguments-renamed
        for encoder in self.encoders:
            try:
                res = encoder(obj)
            except TypeError:
                # Encoders raise TypeError if they cannot deal with it
                continue
            return res

        # Will raise a TypeError
        return super().default(obj)


encode = CleaseEncoder().encode


def object_hook(dct):
    """Wrapper around ASE jsonio object hook"""
    if "__clease_objtype__" in dct:
        objtype = dct.pop("__clease_objtype__")
        return create_clease_object(objtype, dct)
    return aseio.object_hook(dct)


# Note: we disable cyclic imports, as imports here are defered until much later,
# so it's not actually an issue.
# pylint: disable=import-outside-toplevel,cyclic-import
def create_clease_object(objtype, dct):
    if objtype == "concentration":
        from .settings import Concentration

        obj = Concentration.from_dict(dct)
    elif objtype == "basisfunction":
        # It could be multiple types of basis functions
        from .basis_function import basis_function_from_dict

        obj = basis_function_from_dict(dct)
    elif objtype == "ce_settings":
        from .settings import ClusterExpansionSettings

        obj = ClusterExpansionSettings.from_dict(dct)
    elif objtype == "cluster_list":
        from .cluster.cluster_list import ClusterList

        obj = ClusterList.from_dict(dct)
    elif objtype == "cluster":
        from .cluster.cluster import Cluster

        obj = Cluster.from_dict(dct)
    elif objtype == "cluster_fingerprint":
        from .cluster.cluster_fingerprint import ClusterFingerprint

        obj = ClusterFingerprint.from_dict(dct)
    elif objtype == "figure":
        from .datastructures.figure import Figure

        obj = Figure.from_dict(dct)
    elif objtype == "four_vector":
        from .datastructures.four_vector import FourVector

        obj = FourVector.from_dict(dct)
    elif objtype == "system_change":
        from .datastructures.system_changes import SystemChange

        obj = SystemChange.from_dict(dct)
    elif objtype == "mc_step":
        from .datastructures.mc_step import MCStep

        obj = MCStep.from_dict(dct)
    else:
        raise ValueError(f"Cannot load object type: {objtype}")
    assert obj.clease_objtype == objtype
    return obj


cleasedecode = json.JSONDecoder(object_hook=object_hook).decode


def decode(txt, always_array=False):
    obj = cleasedecode(txt)
    obj = aseio.fix_int_keys_in_dicts(obj)
    if always_array:
        obj = aseio.numpyfy(obj)
    return obj


@reader
def read_json(fd, always_array=True):
    dct = decode(fd.read(), always_array=always_array)
    return dct


@writer
def write_json(fd, obj):
    fd.write(encode(obj))


def jsonable(name):
    """Similar to the ASE ase.utils.jsonable decorator, but instead adds a clease_objtype.

    Adds the following methods to the class:
        cls.save
        cls.load
    `load` is added as a classmethod, as it returns a new instance of that class type
    """

    # Define helper methods for saving/loading
    def save_json(self, fd):
        """Method for writing class object to a JSON file."""
        write_json(fd, self)

    @classmethod
    def load_json(cls, fd, **kwargs):
        """Method for loading class object from JSON"""
        obj = read_json(fd, **kwargs)
        assert isinstance(obj, cls)  # pylint: disable=isinstance-second-argument-not-valid-type
        return obj

    def jsonableclass(cls):
        """Wrapper function which adds the save/load methods to the class"""
        assert hasattr(cls, "todict")

        # Define the class attributes
        cls.clease_objtype = name
        cls.save = save_json
        cls.load = load_json
        return cls

    return jsonableclass


class AttrSavable:
    """Mixin for saving simple attrs classes"""

    def todict(self) -> Dict[str, Any]:
        """Convert into a dictionary representation."""
        # Disable recursive, since the json module takes care of that
        return attr.asdict(self, recurse=False)

    @classmethod
    def from_dict(cls, dct: Dict[str, Any]):
        """Load an instance of the class from a dictionary."""
        return cls(**dct)



from clease.datastructures import SystemChanges


class MCConstraint:
    """
    Class for that prevents the MC sampler to run certain moves
    """

    name = "GenericConstraint"

    def __call__(self, system_changes: SystemChanges) -> bool:
        """Return `True` if the trial move is valid.

        Parameters:

        system_changes: list
            System changes. See doc-string of
            `clease.montecarlo.observers.MCObserver`
        """
        return True
